import datetime
import json
import logging

from flask import Blueprint, request, abort, jsonify

from platypush.backend.http.app.utils import authenticate
from platypush.backend.http.app.utils.auth import (
    UserAuthStatus,
    current_user,
    get_current_user_or_auth_status,
)
from platypush.exceptions.user import UserException
from platypush.user import User, UserManager
from platypush.utils import utcnow

auth = Blueprint('auth', __name__)
log = logging.getLogger(__name__)

# Declare routes list
__routes__ = [
    auth,
]


def _dump_session(session, redirect_page='/'):
    return jsonify(
        {
            'status': 'ok',
            'user_id': session.user_id,
            'session_token': session.session_token,
            'expires_at': session.expires_at,
            'redirect': redirect_page,
        }
    )


def _jwt_auth():
    try:
        payload = json.loads(request.get_data(as_text=True))
        username, password = payload['username'], payload['password']
    except Exception:
        log.warning('Invalid payload passed to the auth endpoint')
        abort(400)

    expiry_days = payload.get('expiry_days')
    expires_at = None
    if expiry_days:
        expires_at = datetime.datetime.now() + datetime.timedelta(days=expiry_days)

    user_manager = UserManager()

    try:
        return jsonify(
            {
                'token': user_manager.generate_jwt_token(
                    username=username, password=password, expires_at=expires_at
                ),
            }
        )
    except UserException as e:
        abort(401, str(e))


def _session_auth():
    user_manager = UserManager()
    session_token = request.cookies.get('session_token')
    redirect_page = request.args.get('redirect') or '/'

    if session_token:
        user, session = user_manager.authenticate_user_session(session_token)[:2]
        if user and session:
            return _dump_session(session, redirect_page)

    if request.form:
        username = request.form.get('username')
        password = request.form.get('password')
        code = request.form.get('code')
        remember = request.form.get('remember')
        expires = utcnow() + datetime.timedelta(days=365) if remember else None
        session, status = user_manager.create_user_session(  # type: ignore
            username=username,
            password=password,
            code=code,
            expires_at=expires,
            with_status=True,
        )

        if session:
            return _dump_session(session, redirect_page)

        if status:
            auth_status = UserAuthStatus.by_status(status)
            assert auth_status
            return auth_status.to_response()

    return UserAuthStatus.INVALID_CREDENTIALS.to_response()


def _create_token():
    payload = {}
    try:
        payload = json.loads(request.get_data(as_text=True))
    except json.JSONDecodeError:
        pass

    user = None
    username = payload.get('username')
    password = payload.get('password')
    code = payload.get('code')
    name = payload.get('name')
    expiry_days = payload.get('expiry_days')
    user_manager = UserManager()
    response = get_current_user_or_auth_status(request)

    # Try and authenticate with the credentials passed in the JSON payload
    if username and password:
        user = user_manager.authenticate_user(username, password, code=code)
        if not isinstance(user, User):
            return UserAuthStatus.INVALID_CREDENTIALS.to_response()

    if not user:
        if not (response and isinstance(response, User)):
            return response.to_response()

        user = response

    expires_at = None
    if expiry_days:
        expires_at = datetime.datetime.now() + datetime.timedelta(days=expiry_days)

    try:
        token = UserManager().generate_api_token(
            username=str(user.username), name=name, expires_at=expires_at
        )
        return jsonify({'token': token})
    except UserException:
        return UserAuthStatus.INVALID_CREDENTIALS.to_response()


def _delete_token():
    try:
        payload = json.loads(request.get_data(as_text=True))
        token = payload.get('token')
        assert token
    except (AssertionError, json.JSONDecodeError):
        return UserAuthStatus.INVALID_TOKEN.to_response()

    user_manager = UserManager()

    try:
        token = payload.get('token')
        if not token:
            return UserAuthStatus.INVALID_TOKEN.to_response()

        ret = user_manager.delete_api_token(token)
        if not ret:
            return UserAuthStatus.INVALID_TOKEN.to_response()

        return jsonify({'status': 'ok'})
    except UserException:
        return UserAuthStatus.INVALID_CREDENTIALS.to_response()


def _register_route():
    """Registration endpoint"""
    user_manager = UserManager()
    session_token = request.cookies.get('session_token')
    redirect_page = request.args.get('redirect') or '/'

    if session_token:
        user, session = user_manager.authenticate_user_session(session_token)[:2]
        if user and session:
            return _dump_session(session, redirect_page)

    if user_manager.get_user_count() > 0:
        return UserAuthStatus.REGISTRATION_DISABLED.to_response()

    if not request.form:
        return UserAuthStatus.MISSING_USERNAME.to_response()

    username = request.form.get('username')
    password = request.form.get('password')
    confirm_password = request.form.get('confirm_password')
    remember = request.form.get('remember')

    if not username:
        return UserAuthStatus.MISSING_USERNAME.to_response()
    if not password:
        return UserAuthStatus.MISSING_PASSWORD.to_response()
    if password != confirm_password:
        return UserAuthStatus.PASSWORD_MISMATCH.to_response()

    user_manager.create_user(username=username, password=password)
    session, status = user_manager.create_user_session(  # type: ignore
        username=username,
        password=password,
        expires_at=(utcnow() + datetime.timedelta(days=365) if remember else None),
        with_status=True,
    )

    if session:
        return _dump_session(session, redirect_page)

    if status:
        return status.to_response()  # type: ignore

    return UserAuthStatus.INVALID_CREDENTIALS.to_response()


def _auth_get():
    """
    Get the current authentication status of the user session.
    """
    user_manager = UserManager()
    session_token = request.cookies.get('session_token')
    redirect_page = request.args.get('redirect') or '/'
    user, session, status = user_manager.authenticate_user_session(  # type: ignore
        session_token, with_status=True
    )

    if user and session:
        return _dump_session(session, redirect_page)

    response = get_current_user_or_auth_status(request)
    if isinstance(response, User):
        user = response
        return jsonify(
            {'status': 'ok', 'user_id': user.user_id, 'username': user.username}
        )

    if response:
        status = response

    if status:
        if not isinstance(status, UserAuthStatus):
            status = UserAuthStatus.by_status(status)
        if not status:
            status = UserAuthStatus.INVALID_CREDENTIALS
        return status.to_response()

    return UserAuthStatus.INVALID_CREDENTIALS.to_response()


def _auth_post():
    """
    Authenticate the user session.
    """
    auth_type = request.args.get('type') or 'token'

    if auth_type == 'token':
        return _create_token()

    if auth_type == 'jwt':
        return _jwt_auth()

    if auth_type == 'register':
        return _register_route()

    if auth_type == 'login':
        return _session_auth()

    return UserAuthStatus.INVALID_AUTH_TYPE.to_response()


def _auth_delete():
    """
    Logout/invalidate a token or the current user session.
    """
    # Delete the specified API token if it's passed on the JSON payload
    token = None
    try:
        payload = json.loads(request.get_data(as_text=True))
        token = payload.get('token')
    except json.JSONDecodeError:
        pass

    if token:
        return _delete_token()

    user_manager = UserManager()
    session_token = request.cookies.get('session_token')
    redirect_page = request.args.get('redirect') or '/'

    if session_token:
        user, session = user_manager.authenticate_user_session(session_token)[:2]
        if user and session:
            user_manager.delete_user_session(session_token)
            return jsonify({'status': 'ok', 'redirect': redirect_page})

    return UserAuthStatus.INVALID_SESSION.to_response()


def _tokens_get():
    user = current_user()
    if not user:
        return UserAuthStatus.INVALID_CREDENTIALS.to_response()

    tokens = UserManager().get_api_tokens(username=str(user.username))
    return jsonify(
        {
            'tokens': [
                {
                    'id': t.id,
                    'name': t.name,
                    'created_at': t.created_at,
                    'expires_at': t.expires_at,
                }
                for t in tokens
            ]
        }
    )


def _tokens_delete():
    args = {}

    try:
        payload = json.loads(request.get_data(as_text=True))
        token = payload.get('token')
        if token:
            args['token'] = token
        else:
            token_id = payload.get('token_id')
            if token_id:
                args['token_id'] = token_id

        assert args, 'No token or token_id specified'
    except (AssertionError, json.JSONDecodeError):
        return UserAuthStatus.INVALID_TOKEN.to_response()

    user_manager = UserManager()
    user = current_user()
    if not user:
        return UserAuthStatus.INVALID_CREDENTIALS.to_response()

    args['username'] = str(user.username)

    try:
        user_manager.delete_api_token(**args)
        return jsonify({'status': 'ok'})
    except AssertionError as e:
        return (
            jsonify({'status': 'error', 'error': 'bad_request', 'message': str(e)}),
            400,
        )
    except UserException:
        return UserAuthStatus.INVALID_CREDENTIALS.to_response()
    except Exception as e:
        log.error('Token deletion error', exc_info=e)

    return UserAuthStatus.UNKNOWN_ERROR.to_response()


@auth.route('/auth', methods=['GET', 'POST', 'DELETE'])
def auth_endpoint():
    """
    Authentication endpoint. It validates the user credentials provided over a
    JSON payload with the following structure:

        .. code-block:: json

            {
                "username": "USERNAME",
                "password": "PASSWORD",
                "code": "2FA_CODE",
                "expiry_days": "The generated token should be valid for these many days"
            }

    ``expiry_days`` is optional, and if omitted or set to zero the token will
    be valid indefinitely.

    Upon successful validation, a new JWT token will be generated using the
    service's self-generated RSA key-pair and it will be returned to the user.
    The token can then be used to authenticate API calls to ``/execute`` by
    setting the ``Authorization: Bearer <TOKEN_HERE>`` header upon HTTP calls.

    :return: Return structure:

        .. code-block:: json

            {
                "token": "<generated token here>"
            }
    """
    if request.method == 'GET':
        return _auth_get()

    if request.method == 'POST':
        return _auth_post()

    if request.method == 'DELETE':
        return _auth_delete()

    return UserAuthStatus.INVALID_METHOD.to_response()


@auth.route('/tokens', methods=['GET', 'DELETE'])
@authenticate()
def tokens_route():
    """
    :return: The list of API tokens created by the logged in user.
        Note that this endpoint is only accessible by authenticated users
        and it won't return the clear-text token values, as those aren't
        stored in the database anyway.
    """
    if request.method == 'GET':
        return _tokens_get()

    if request.method == 'DELETE':
        return _tokens_delete()

    return UserAuthStatus.INVALID_METHOD.to_response()


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
from typing import Callable, Dict, Union

from platypush.config import Config
from . import EntityManager


class ProcedureEntityManager(EntityManager, ABC):
    """
    Base class for integrations that can run and manage procedures.
    """

    @abstractmethod
    def exec(self, procedure: str, *args, **kwargs):
        """
        Run a procedure.

        :param procedure: Procedure to run, by name.
        :param args: Arguments to pass to the procedure.
        :param kwargs: Keyword arguments to pass to the procedure.
        """
        raise NotImplementedError()

    @property
    def _all_procedures(self) -> Dict[str, Union[dict, Callable]]:
        """
        :return: All the procedures that can be run by this entity manager.
        """
        return Config.get_procedures()



from platypush.context import get_plugin
from platypush.plugins.media._search import MediaSearcher


# pylint: disable=too-few-public-methods
class PlexMediaSearcher(MediaSearcher):
    """
    Plex media searcher.
    """

    def supports(self, type: str) -> bool:
        return type == 'plex'

    def search(self, query: str, *_, **__):
        """
        Performs a Plex search using the configured :class:`platypush.plugins.media.plex.MediaPlexPlugin` instance if
        it is available.
        """

        try:
            plex = get_plugin('media.plex')
        except RuntimeError:
            return []

        assert plex, 'No Plex plugin configured'
        self.logger.info('Searching Plex for "%s"', query)
        results = []

        for result in plex.search(title=query).output:
            results.extend(self._flatten_result(result))

        self.logger.info(
            '%d Plex results found for the search query "%s"', len(results), query
        )
        return results

    @staticmethod
    def _flatten_result(result):
        def parse_part(media, part, episode=None, sub_media=None):
            if 'episodes' in media:
                del media['episodes']

            return {
                **{k: v for k, v in result.items() if k not in ['media', 'type']},
                'media_type': result.get('type'),
                'type': 'plex',
                **{k: v for k, v in media.items() if k not in ['parts']},
                **part,
                'title': (
                    result.get('title', '')
                    + (
                        ' [' + (episode or {}).get('season_episode', '') + ']'
                        if (episode or {}).get('season_episode')
                        else ''
                    )
                    + (
                        ' ' + (sub_media or {}).get('title', '')
                        if (sub_media or {}).get('title')
                        else ''
                    ),
                ),
                'summary': (
                    (episode or {}).get('summary')
                    if (episode or {}).get('summary')
                    else media.get('summary')
                ),
            }

        results = []

        for media in result.get('media', []):
            if 'episodes' in media:
                for episode in media['episodes']:
                    for sub_media in episode.get('media', []):
                        for part in sub_media.get('parts', []):
                            results.append(
                                parse_part(
                                    media=media,
                                    episode=episode,
                                    sub_media=sub_media,
                                    part=part,
                                )
                            )
            else:
                for part in media.get('parts', []):
                    results.append(parse_part(media=media, part=part))

        return results


# vim:sw=4:ts=4:et:



import os
import queue
from typing import Optional

from platypush.context import get_plugin

from .. import FileMediaResource
from ._base import MediaResourceParser


# pylint: disable=too-few-public-methods
class TorrentResourceParser(MediaResourceParser):
    """
    Parser for magnet links.
    """

    @staticmethod
    def _torrent_event_handler(evt_queue):
        def handler(event):
            # More than 5% of the torrent has been downloaded
            if event.args.get('progress', 0) > 5 and event.args.get('files'):
                evt_queue.put(
                    [f for f in event.args['files'] if f.is_media_file(f.filename)]
                )

        return handler

    def parse(self, resource: str, *_, **__) -> Optional[FileMediaResource]:
        if not resource.startswith('magnet:?'):
            return None

        torrents = get_plugin(self._media.torrent_plugin)
        assert torrents, f'{self._media.torrent_plugin} plugin not configured'

        evt_queue = queue.Queue()
        torrents.download(
            resource,
            download_dir=self._media.download_dir,
            _async=True,
            is_media=True,
            event_hndl=self._torrent_event_handler(evt_queue),
        )

        resources = [f for f in evt_queue.get()]  # noqa: C416,R1721

        if resources:
            self._media._videos_queue = videos_queue = sorted(resources)
            resource = videos_queue.pop(0)
        else:
            raise RuntimeError(f'No media file found in torrent {resource}')

        assert resource, 'Unable to find any compatible media resource'
        return FileMediaResource(
            resource=resource,
            url=f'file://{resource}',
            media_plugin=self._media,
            title=os.path.basename(resource),
            filename=os.path.basename(resource),
        )


# vim:sw=4:ts=4:et:



from marshmallow import fields, pre_load

from .._base import SystemBaseSchema


class ConnectionSchema(SystemBaseSchema):
    """
    Base schema for connections.
    """

    fd = fields.Int(
        required=True,
        metadata={
            'description': 'File descriptor number.',
            'example': 3,
        },
    )

    family = fields.String(
        allow_none=True,
        metadata={
            'description': 'Address family.',
            'example': 'AF_INET',
        },
    )

    type = fields.String(
        allow_none=True,
        metadata={
            'description': 'Socket type.',
            'example': 'SOCK_STREAM',
        },
    )

    local_address = fields.String(
        allow_none=True,
        metadata={
            'description': 'Local address.',
            'example': '192.168.1.3',
        },
    )

    local_port = fields.Int(
        allow_none=True,
        metadata={
            'description': 'Local port.',
            'example': 1234,
        },
    )

    remote_address = fields.String(
        allow_none=True,
        metadata={
            'description': 'Remote address.',
            'example': '192.168.1.4',
        },
    )

    remote_port = fields.Int(
        allow_none=True,
        metadata={
            'description': 'Remote port.',
            'example': 5678,
        },
    )

    status = fields.String(
        allow_none=True,
        metadata={
            'description': 'Connection status.',
            'example': 'ESTABLISHED',
        },
    )

    pid = fields.Int(
        allow_none=True,
        metadata={
            'description': 'ID of the process that owns the connection.',
            'example': 1234,
        },
    )

    @pre_load
    def pre_load(self, data, **_) -> dict:
        data = super().pre_load(data)
        addr_mapping = {
            'laddr': ('local_address', 'local_port'),
            'raddr': ('remote_address', 'remote_port'),
        }

        # Parse laddr/raddr attributes
        for ext_attr, (addr_attr, port_attr) in addr_mapping.items():
            value = data.pop(ext_attr, None)
            if not value:
                data[addr_attr] = data[port_attr] = None
            elif isinstance(value, tuple):
                data[addr_attr], data[port_attr] = value
            elif isinstance(value, str):
                data[addr_attr] = value
                data[port_attr] = None

        # Handle enum values
        for attr in ['type', 'family']:
            value = data.pop(attr, None)
            if value is not None:
                data[attr] = value.name

        if data.get('status') == 'NONE':
            data['status'] = None

        return data



import enum
import logging
import re
from copy import deepcopy
from dataclasses import dataclass, field
from functools import wraps

from queue import LifoQueue
from typing import Any, Dict, Iterable, List, Optional

from ..common import exec_wrapper
from ..config import Config
from ..message.request import Request
from ..message.response import Response

logger = logging.getLogger('platypush')


class StatementType(enum.Enum):
    """
    Enumerates the possible statements in a procedure.
    """

    BREAK = 'break'
    CONTINUE = 'continue'
    RETURN = 'return'
    SET = 'set'


@dataclass
class Statement:
    """
    Models a statement in a procedure.
    """

    type: StatementType
    argument: Optional[Any] = None

    @classmethod
    def build(cls, statement: str):
        """
        Builds a statement from a string.
        """

        m = re.match(r'\s*return\s*(.*)\s*', statement, re.IGNORECASE)
        if m:
            return ReturnStatement(argument=m.group(1))

        return cls(StatementType(statement.lower()))

    def run(self, *_, **__) -> Optional[Any]:
        """
        Executes the statement.
        """


@dataclass
class ReturnStatement(Statement):
    """
    Models a return statement in a procedure.
    """

    type: StatementType = StatementType.RETURN

    def run(self, *_, **context) -> Any:
        return Response(
            output=Request.expand_value_from_context(
                self.argument, **_update_context(context)
            )
        )


@dataclass
class SetStatement(Statement):
    """
    Models a set variable statement in a procedure.
    """

    type: StatementType = StatementType.SET
    vars: dict = field(default_factory=dict)

    def run(self, *_, **context):
        vars = deepcopy(self.vars)  # pylint: disable=redefined-builtin
        for k, v in vars.items():
            vars[k] = Request.expand_value_from_context(v, **context)

        context.update(vars)
        return Response(output=vars)


class Procedure:
    """Procedure class. A procedure is a pre-configured list of requests"""

    def __init__(self, name, _async, requests, args=None, backend=None):
        """
        Params:
            name     -- Procedure name
            _async   -- Whether the actions in the procedure are supposed to
                        be executed sequentially or in parallel (True or False)
            requests -- List of platypush.message.request.Request objects
        """

        self.name = name
        self._async = _async
        self.requests = requests
        self.backend = backend
        self.args = args or {}
        self._should_return = False

        for req in requests:
            req.backend = self.backend

    @classmethod
    # pylint: disable=too-many-branches,too-many-statements
    def build(
        cls,
        name,
        _async,
        requests,
        args=None,
        backend=None,
        procedure_class=None,
        **kwargs,
    ):
        reqs = []
        for_count = 0
        while_count = 0
        if_count = 0
        if_config = LifoQueue()
        procedure_class = procedure_class or cls
        key = None
        kwargs.pop('id', None)

        for request_config in requests:
            # Check if it's a break/continue/return statement
            if isinstance(request_config, str):
                cls._flush_if_statements(reqs, if_config)
                reqs.append(Statement.build(request_config))
                continue

            # Check if it's a return statement with a value
            if (
                len(request_config.keys()) == 1
                and list(request_config.keys())[0] == StatementType.RETURN.value
            ):
                cls._flush_if_statements(reqs, if_config)
                reqs.append(
                    ReturnStatement(argument=request_config[StatementType.RETURN.value])
                )
                continue

            # Check if it's a variable set statement
            if (len(request_config.keys()) == 1) and (
                list(request_config.keys())[0] == StatementType.SET.value
            ):
                cls._flush_if_statements(reqs, if_config)
                reqs.append(SetStatement(vars=request_config[StatementType.SET.value]))
                continue

            # Check if this request is an if-else
            if len(request_config.keys()) >= 1:
                key = list(request_config.keys())[0]
                m = re.match(r'\s*(if)\s+\${(.*)}\s*', key)

                if m:
                    cls._flush_if_statements(reqs, if_config)
                    if_count += 1
                    if_name = f'{name}__if_{if_count}'
                    condition = m.group(2)

                    if_config.put(
                        {
                            'name': if_name,
                            '_async': False,
                            'requests': request_config[key],
                            'condition': condition,
                            'else_branch': [],
                            'backend': backend,
                        }
                    )

                    continue

                if key == 'else':
                    if if_config.empty():
                        raise RuntimeError(
                            f'else statement with no associated if in {name}'
                        )

                    conf = if_config.get()
                    conf['else_branch'] = request_config[key]
                    if_config.put(conf)

            if not if_config.empty():
                reqs.append(IfProcedure.build(**(if_config.get())))
                if key == 'else':
                    continue

            # Check if this request is a for loop
            if len(request_config.keys()) == 1:
                key = list(request_config.keys())[0]
                m = re.match(r'\s*(fork?)\s+([\w\d_]+)\s+in\s+(.*)\s*', key)

                if m:
                    for_count += 1
                    loop_name = f'{name}__for_{for_count}'

                    # A 'for' loop is synchronous. Declare a 'fork' loop if you
                    # want to process the elements in the iterable in parallel
                    _async = m.group(1) == 'fork'
                    iterator_name = m.group(2)
                    iterable = m.group(3)

                    loop = ForProcedure.build(
                        name=loop_name,
                        _async=_async,
                        requests=request_config[key],
                        backend=backend,
                        iterator_name=iterator_name,
                        iterable=iterable,
                    )

                    reqs.append(loop)
                    continue

            # Check if this request is a while loop
            if len(request_config.keys()) == 1:
                key = list(request_config.keys())[0]
                m = re.match(r'\s*while\s+\${(.*)}\s*', key)

                if m:
                    while_count += 1
                    loop_name = f'{name}__while_{while_count}'
                    condition = m.group(1).strip()

                    loop = WhileProcedure.build(
                        name=loop_name,
                        _async=False,
                        requests=request_config[key],
                        condition=condition,
                        backend=backend,
                    )

                    reqs.append(loop)
                    continue

            request_config['origin'] = Config.get('device_id')
            if 'target' not in request_config:
                request_config['target'] = request_config['origin']

            request = Request.build(request_config)
            reqs.append(request)

        cls._flush_if_statements(reqs, if_config)

        return procedure_class(
            name=name,
            _async=_async,
            requests=reqs,
            args=args,
            backend=backend,
            **kwargs,
        )

    @staticmethod
    def _flush_if_statements(requests: List, if_config: LifoQueue):
        while not if_config.empty():
            pending_if = if_config.get()
            requests.append(IfProcedure.build(**pending_if))

    # pylint: disable=too-many-branches,too-many-statements
    def execute(
        self,
        n_tries: int = 1,
        __stack__: Optional[Iterable] = None,
        new_context: Optional[Dict[str, Any]] = None,
        **context,
    ):
        """
        Execute the requests in the procedure.

        :param n_tries: Number of tries in case of failure before raising a RuntimeError.
        """
        __stack__ = (self,) if not __stack__ else (self, *__stack__)
        new_context = new_context or {}

        if self.args:
            args = self.args.copy()
            for k, v in args.items():
                args[k] = context[k] = Request.expand_value_from_context(v, **context)
            logger.info('Executing procedure %s with arguments %s', self.name, args)
        else:
            logger.info('Executing procedure %s', self.name)

        response = Response()
        token = Config.get('token')
        context = _update_context(context)
        locals().update(context)

        # pylint: disable=too-many-nested-blocks
        for request in self.requests:
            if callable(request):
                response = request(**context)
                continue

            context['_async'] = self._async
            context['n_tries'] = n_tries
            context['__stack__'] = __stack__
            context['new_context'] = new_context

            if isinstance(request, Statement):
                if isinstance(request, ReturnStatement):
                    response = request.run(**context)
                    self._should_return = True
                    for proc in __stack__:
                        proc._should_return = True  # pylint: disable=protected-access

                    break

                if isinstance(request, SetStatement):
                    rs: dict = request.run(**context).output  # type: ignore
                    context.update(rs)
                    new_context.update(rs)
                    locals().update(rs)
                    continue

                if request.type in [StatementType.BREAK, StatementType.CONTINUE]:
                    for proc in __stack__:
                        if isinstance(proc, LoopProcedure):
                            if request.type == StatementType.BREAK:
                                setattr(proc, '_should_break', True)  # noqa: B010
                            else:
                                setattr(proc, '_should_continue', True)  # noqa: B010
                            break

                        proc._should_return = True  # pylint: disable=protected-access

                    break

            should_continue = getattr(self, '_should_continue', False)
            should_break = getattr(self, '_should_break', False)
            if self._should_return or should_continue or should_break:
                break

            if token and not isinstance(request, Statement):
                request.token = token

            exec_ = getattr(request, 'execute', None)
            if callable(exec_):
                response = exec_(**context)
                context.update(context.get('new_context', {}))

            if not self._async and response:
                if isinstance(response.output, dict):
                    context.update(response.output)

                context['output'] = response.output
                context['errors'] = response.errors
                new_context.update(context)
                locals().update(context)

            if self._should_return:
                break

        return response or Response()

    def to_dict(self):
        return {
            'name': self.name,
            'requests': self.requests,
            'args': self.args,
            '_async': self._async,
        }


class LoopProcedure(Procedure):
    """
    Base class while and for/fork loops.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._should_break = False
        self._should_continue = False


class ForProcedure(LoopProcedure):
    """
    Models a loop procedure, which expresses a construct similar to a
    for loop in a programming language. The 'for' keyword implies a synchronous
    loop, i.e. the nested actions will be executed in sequence. Use 'fork'
    instead of 'for' if you want to run the actions in parallel.

    Example::

        procedure.sync.process_results:
            - action: http.get
              args:
                  url: https://some-service/some/json/endpoint
                  # Example response: { "results": [ {"id":1, "name":"foo"}, {"id":2,"name":"bar"} ]}
            - for result in ${results}:
                - action: some.custom.action
                  args:
                      id: ${result['id']}
                      name: ${result['name']}

    """

    def __init__(
        self,
        name,
        iterator_name,
        iterable,
        requests,
        _async=False,
        args=None,
        backend=None,
    ):
        super().__init__(
            name=name, _async=_async, requests=requests, args=args, backend=backend
        )
        self.iterator_name = iterator_name
        self.iterable = iterable

    # pylint: disable=eval-used
    def execute(self, *_, **context):
        ctx = _update_context(context)
        locals().update(ctx)

        try:
            iterable = eval(self.iterable)
            assert hasattr(
                iterable, '__iter__'
            ), f'Object of type {type(iterable)} is not iterable: {iterable}'
        except Exception as e:
            logger.debug('Iterable %s expansion error: %s', self.iterable, e)
            iterable = Request.expand_value_from_context(self.iterable, **ctx)

        response = Response()

        for item in iterable:
            ctx[self.iterator_name] = item
            response = super().execute(**ctx)
            ctx.update(ctx.get('new_context', {}))

            if response.output and isinstance(response.output, dict):
                ctx = _update_context(ctx, **response.output)

            if self._should_return:
                logger.info('Returning from %s', self.name)
                break

            if self._should_continue:
                self._should_continue = False
                logger.info('Continuing loop %s', self.name)
                continue

            if self._should_break:
                self._should_break = False
                logger.info('Breaking loop %s', self.name)
                break

        return response


class WhileProcedure(LoopProcedure):
    """
    Models a while loop procedure.

    Example::

        procedure.process_results:
            - action: http.get
              args:
                  url: https://some-service/some/json/endpoint
                  # Example response: {"id":1, "name":"foo"}}

            - while ${output}:
                - action: some.custom.action
                  args:
                      id: ${id}
                      name: ${name}

                - action: http.get
                  args:
                      url: https://some-service/some/json/endpoint
                      # Example response: {"id":1, "name":"foo"}}

    """

    def __init__(
        self, name, condition, requests, _async=False, args=None, backend=None
    ):
        super().__init__(
            name=name, _async=_async, requests=requests, args=args, backend=backend
        )
        self.condition = condition

    def execute(self, *_, **context):
        response = Response()
        ctx = _update_context(context)
        locals().update(ctx)

        while True:
            condition_true = eval(self.condition)  # pylint: disable=eval-used
            if not condition_true:
                break

            response = super().execute(**ctx)
            ctx.update(ctx.get('new_context', {}))
            if response.output and isinstance(response.output, dict):
                _update_context(ctx, **response.output)

            locals().update(ctx)

            if self._should_return:
                logger.info('Returning from %s', self.name)
                break

            if self._should_continue:
                self._should_continue = False
                logger.info('Continuing loop %s', self.name)
                continue

            if self._should_break:
                self._should_break = False
                logger.info('Breaking loop %s', self.name)
                break

        return response


class IfProcedure(Procedure):
    """
    Models an if-else construct.

    Example::

        procedure.sync.process_results:
            - action: http.get
              args:
                url: https://some-service/some/json/endpoint
                # Example response: { "sensors": [ {"temperature": 18 } ] }

            - if ${sensors['temperature'] < 20}:
              - action: shell.exec
                args:
                  cmd: '/path/turn_on_heating.sh'
            - else:
              - action: shell.exec
                args:
                  cmd: '/path/turn_off_heating.sh'

    """

    def __init__(
        self,
        name,
        condition,
        requests,
        else_branch=None,
        args=None,
        backend=None,
        id=None,  # pylint: disable=redefined-builtin
        **kwargs,
    ):
        kwargs['_async'] = False
        self.condition = condition
        self.else_branch = else_branch
        reqs = []

        for req in requests:
            if isinstance(req, dict):
                req['origin'] = Config.get('device_id')
                req['id'] = id
                if 'target' not in req:
                    req['target'] = req['origin']

                req = Request.build(req)

            reqs.append(req)

        super().__init__(name=name, requests=reqs, args=args, backend=backend, **kwargs)

    @classmethod
    # pylint: disable=arguments-differ
    def build(
        cls,
        name,
        *_,
        condition,
        requests,
        args=None,
        backend=None,
        id=None,  # pylint: disable=redefined-builtin
        else_branch=None,
        **kwargs,
    ):
        kwargs['_async'] = False
        if else_branch:
            else_branch = super().build(
                name=name + '__else',
                requests=else_branch,
                args=args,
                backend=backend,
                id=id,
                procedure_class=Procedure,
                **kwargs,
            )

        return super().build(
            name=name,
            condition=condition,
            requests=requests,
            else_branch=else_branch,
            args=args,
            backend=backend,
            id=id,
            **kwargs,
        )

    def execute(self, *_, **context):
        ctx = _update_context(context)
        locals().update(ctx)
        condition_true = eval(self.condition)  # pylint: disable=eval-used
        response = Response()

        if condition_true:
            response = super().execute(**ctx)
        elif self.else_branch:
            response = self.else_branch.execute(**ctx)

        return response


def _update_context(context: Optional[Dict[str, Any]] = None, **kwargs):
    ctx = context or {}
    ctx = {**ctx.get('context', {}), **ctx, **kwargs}
    for k, v in ctx.items():
        ctx[k] = Request.expand_value_from_context(v, **ctx)

    return ctx


def procedure(name_or_func: Optional[str] = None, *upper_args, **upper_kwargs):
    name = name_or_func if isinstance(name_or_func, str) else None

    def func_wrapper(f):
        """
        Public decorator to mark a function as a procedure.
        """
        import inspect

        f.procedure = True
        f.procedure_name = name
        f._source = inspect.getsourcefile(f)  # pylint: disable=protected-access
        f._line = inspect.getsourcelines(f)[1]  # pylint: disable=protected-access

        @wraps(f)
        def _execute_procedure(*args, **kwargs):
            args = [*upper_args, *args]
            kwargs = {**upper_kwargs, **kwargs}
            return exec_wrapper(f, *args, **kwargs)

        return _execute_procedure

    if callable(name_or_func):
        return func_wrapper(name_or_func)

    return func_wrapper


# vim:sw=4:ts=4:et:



import enum
from dataclasses import dataclass, field
from typing import Optional


class PlayerState(enum.Enum):
    """
    Models the possible states of a media player
    """

    STOP = 'stop'
    PLAY = 'play'
    PAUSE = 'pause'
    IDLE = 'idle'


class DownloadState(enum.Enum):
    """
    Enum that represents the status of a download.
    """

    IDLE = 'idle'
    STARTED = 'started'
    DOWNLOADING = 'downloading'
    PAUSED = 'paused'
    COMPLETED = 'completed'
    CANCELLED = 'cancelled'
    ERROR = 'error'


@dataclass
class MediaDirectoryIcon:
    """
    Dataclass that represents a media directory icon.
    """

    class_: str = 'fas fa-folder'
    url: Optional[str] = None

    def to_dict(self) -> dict:
        """
        Convert the MediaDirectoryIcon instance to a dictionary.
        """
        return {'class': self.class_, 'url': self.url}


@dataclass
class MediaDirectory:
    """
    Dataclass that represents a media directory.
    """

    name: str
    path: str
    icon: MediaDirectoryIcon = field(default_factory=MediaDirectoryIcon)

    @classmethod
    def build(
        cls,
        name: str,
        path: str,
        icon_class: Optional[str] = None,
        icon_url: Optional[str] = None,
    ) -> 'MediaDirectory':
        """
        Create a MediaDirectory instance from a dictionary.
        """
        icon_class = icon_class or 'fas fa-folder'
        return cls(name, path, MediaDirectoryIcon(icon_class, icon_url))

    def to_dict(self) -> dict:
        """
        Convert the MediaDirectory instance to a dictionary.
        """
        return {'name': self.name, 'path': self.path, 'icon': self.icon.to_dict()}


# vim:sw=4:ts=4:et:



import json
import signal
import subprocess
import threading
from contextlib import suppress
from typing import Optional

from platypush.plugins.media._model import DownloadState
from platypush.plugins.media._resource import MediaResource, YoutubeMediaResource
from platypush.utils import wait_for_either

from ._base import DownloadThread, MediaResourceDownloader


class YoutubeResourceDownloader(MediaResourceDownloader):
    """
    Downloader for YouTube URLs.
    """

    def supports(self, resource_type: MediaResource) -> bool:
        return isinstance(resource_type, YoutubeMediaResource)

    def download(  # type: ignore[override]
        self,
        resource: YoutubeMediaResource,
        path: str,
        youtube_format: Optional[str] = None,
        youtube_audio_format: Optional[str] = None,
        merge_output_format: Optional[str] = None,
        only_audio: bool = False,
        **_,
    ) -> 'YouTubeDownloadThread':
        path = self.get_download_path(
            resource=resource, directory=path, youtube_format=youtube_format
        )

        download_thread = YouTubeDownloadThread(
            url=resource.url,
            path=path,
            ytdl=self._media._ytdl,  # pylint: disable=protected-access
            only_audio=only_audio,
            youtube_format=youtube_format or self._media.youtube_format,
            youtube_audio_format=youtube_audio_format
            or self._media.youtube_audio_format,
            merge_output_format=merge_output_format or self._media.merge_output_format,
            on_start=self._media.on_download_start,
            post_event=self._media.post_event,
            stop_event=self._media._should_stop,  # pylint: disable=protected-access
        )

        self._media.start_download(download_thread)
        return download_thread

    def get_download_path(
        self,
        resource: YoutubeMediaResource,
        *_,
        directory: Optional[str] = None,
        filename: Optional[str] = None,
        youtube_format: Optional[str] = None,
        **__,
    ) -> str:
        youtube_format = youtube_format or self._media.youtube_format

        if not filename:
            filename = resource.filename

        if not filename:
            with subprocess.Popen(
                [
                    self._media._ytdl,  # pylint: disable=protected-access
                    *(
                        [
                            '-f',
                            youtube_format,
                        ]
                        if youtube_format
                        else []
                    ),
                    '-O',
                    '%(title)s.%(ext)s',
                    resource.url,
                ],
                stdout=subprocess.PIPE,
            ) as proc:
                assert proc.stdout, 'yt-dlp stdout is None'
                filename = proc.stdout.read().decode()[:-1]

        return super().get_download_path(
            resource, directory=directory, filename=filename
        )

    def supports_only_audio(self) -> bool:
        return True


class YouTubeDownloadThread(DownloadThread):
    """
    Thread that downloads a YouTube URL to a file.
    """

    def __init__(
        self,
        *args,
        ytdl: str,
        youtube_format: Optional[str] = None,
        youtube_audio_format: Optional[str] = None,
        merge_output_format: Optional[str] = None,
        only_audio: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._ytdl = ytdl
        self._youtube_format = youtube_format
        self._youtube_audio_format = youtube_audio_format
        self._merge_output_format = merge_output_format
        self._only_audio = only_audio
        self._proc = None
        self._proc_lock = threading.Lock()

    def _parse_progress(self, line: str):
        try:
            progress = json.loads(line)
        except json.JSONDecodeError:
            return

        status = progress.get('status')
        if not status:
            return

        if status == 'finished':
            self.progress = 100
            return

        if status == 'paused':
            self.state = DownloadState.PAUSED
        elif status == 'downloading':
            self.state = DownloadState.DOWNLOADING

        self.size = int(progress.get('total_bytes_estimate', 0)) or self.size
        if self.size:
            downloaded = int(progress.get('downloaded_bytes', 0))
            self.progress = (downloaded / self.size) * 100

    def _run(self):
        youtube_format = (
            self._youtube_audio_format if self._only_audio else self._youtube_format
        )

        ytdl_cmd = [
            self._ytdl,
            '--newline',
            '--progress',
            '--progress-delta',
            str(self._progress_update_interval),
            '--progress-template',
            '%(progress)j',
            *(['-x'] if self._only_audio else []),
            *(['-f', youtube_format] if youtube_format else []),
            self.url,
            '-o',
            self.path,
        ]

        self.logger.info('Executing command %r', ytdl_cmd)
        err = None

        with subprocess.Popen(ytdl_cmd, stdout=subprocess.PIPE) as self._proc:
            if self._proc.stdout:
                for line in self._proc.stdout:
                    self.logger.debug(
                        '%s output: %s', self._ytdl, line.decode().strip()
                    )

                    self._parse_progress(line.decode())

                    if self.should_stop():
                        self.stop()
                        return self._proc.returncode == 0

                    if self._paused.is_set():
                        wait_for_either(self._downloading, self._stop_event)

        if self._proc.returncode != 0:
            err = self._proc.stderr.read().decode() if self._proc.stderr else None
            raise RuntimeError(
                f'{self._ytdl} failed with return code {self._proc.returncode}: {err}'
            )

        return True

    def pause(self):
        with self._proc_lock:
            if self._proc:
                self._proc.send_signal(signal.SIGSTOP)

        super().pause()

    def resume(self):
        with self._proc_lock:
            if self._proc:
                self._proc.send_signal(signal.SIGCONT)

        super().resume()

    def stop(self):
        state = None

        with suppress(IOError, OSError), self._proc_lock:
            if self._proc:
                if self._proc.poll() is None:
                    self._proc.terminate()
                    self._proc.wait(timeout=3)
                    if self._proc.returncode is None:
                        self._proc.kill()

                    state = DownloadState.CANCELLED
                elif self._proc.returncode != 0:
                    state = DownloadState.ERROR
                else:
                    state = DownloadState.COMPLETED

                self._proc = None

        super().stop()

        if state:
            self.state = state

    def on_close(self):
        self.stop()
        super().on_close()



import os
import subprocess
from dataclasses import dataclass
from typing import IO, Optional, Sequence

from ._base import PopenMediaResource


@dataclass
class YoutubeMediaResource(PopenMediaResource):
    """
    Models a YouTube media resource.
    """

    id: Optional[str] = None
    is_temporary: bool = False

    def _generate_file(self, merge_format: str):
        self.resource = os.path.join(
            self._media.cache_dir, f'platypush-yt-dlp-{self.id}.{merge_format}'
        )
        self.is_temporary = True

    def _prepare_file(self, merge_output_format: str):
        if not self.resource:
            self._generate_file(merge_output_format)

        filename = (
            self.resource[len('file://') :]
            if self.resource.startswith('file://')
            else self.resource
        )

        # Remove the file if it already exists and it's empty, to avoid YTDL
        # errors
        if (
            os.path.exists(os.path.abspath(filename))
            and os.path.getsize(os.path.abspath(filename)) == 0
        ):
            self._logger.debug('Removing empty file: %s', filename)
            os.unlink(os.path.abspath(filename))

    def open(
        self,
        *args,
        youtube_format: Optional[str] = None,
        merge_output_format: Optional[str] = None,
        cache_streams: bool = False,
        ytdl_args: Optional[Sequence[str]] = None,
        **kwargs,
    ) -> IO:
        if self.proc is None:
            merge_output_format = merge_output_format or self._media.merge_output_format
            use_file = (
                not self._media.supports_local_pipe or cache_streams or self.resource
            )

            if use_file:
                self._prepare_file(merge_output_format=merge_output_format)

            output = ['-o', self.resource] if use_file else ['-o', '-']
            youtube_format = youtube_format or self._media.youtube_format
            ytdl_args = ytdl_args or self._media.ytdl_args
            cmd = [
                self._media._ytdl,  # pylint: disable=protected-access
                '--no-part',
                *(
                    [
                        '-f',
                        youtube_format,
                    ]
                    if youtube_format
                    else []
                ),
                *(
                    ['--merge-output-format', merge_output_format]
                    if merge_output_format
                    else []
                ),
                *output,
                *ytdl_args,
                self.url,
            ]

            proc_args = {}
            if not use_file:
                proc_args['stdout'] = subprocess.PIPE

            self._logger.debug('Running command: %s', ' '.join(cmd))
            self._logger.debug('Media resource: %s', self.to_dict())
            self.proc = subprocess.Popen(  # pylint: disable=consider-using-with
                cmd, **proc_args
            )

            if use_file:
                self._wait_for_download_start()
                self.fd = open(  # pylint: disable=consider-using-with
                    self.resource, 'rb'
                )
            elif self.proc.stdout:
                self.fd = self.proc.stdout

        return super().open(*args, **kwargs)

    def _wait_for_download_start(self) -> None:
        self._logger.info('Waiting for download to start on file: %s', self.resource)

        while True:
            file = self.resource
            if not file:
                self._logger.info('No file found to wait for download')
                break

            if not self.proc:
                self._logger.info('No download process found to wait')
                break

            if self.proc.poll() is not None:
                self._logger.info(
                    'Download process exited with status %d', self.proc.returncode
                )
                break

            # The file must exist and be at least 5MB in size
            if os.path.exists(file) and os.path.getsize(file) > 5 * 1024 * 1024:
                self._logger.info('Download started, process PID: %s', self.proc.pid)
                break

            try:
                self.proc.wait(1)
            except subprocess.TimeoutExpired:
                pass

    def close(self) -> None:
        super().close()

        if self.is_temporary and self.resource and os.path.exists(self.resource):
            try:
                self._logger.debug('Removing temporary file: %s', self.resource)
                os.unlink(self.resource)
            except FileNotFoundError:
                pass


# vim:sw=4:ts=4:et:



from dataclasses import dataclass
from typing import Optional


@dataclass
class Battery:
    """
    System battery sensor representation.
    """

    seconds_left: Optional[float] = None
    power_plugged: Optional[bool] = None
    value: Optional[float] = None



from marshmallow import fields, pre_load

from platypush.schemas.dataclasses import percent_field
from .._base import SystemBaseSchema


class DiskSchema(SystemBaseSchema):
    """
    Base schema for disk stats.
    """

    device = fields.String(
        required=True,
        metadata={
            'description': 'Path/identifier of the disk/partition',
            'example': '/dev/sda1',
        },
    )

    mountpoint = fields.String(
        metadata={
            'description': 'Mountpoint of the disk/partition',
            'example': '/mnt/data',
        }
    )

    fstype = fields.String(
        metadata={
            'description': 'Filesystem type',
            'example': 'ext4',
        }
    )

    opts = fields.String(  # type: ignore
        metadata={
            'description': 'Mount options',
            'example': 'rw,relatime',
        }
    )

    total = fields.Integer(
        metadata={
            'description': 'Total disk space in bytes',
            'example': 1024**3,
        }
    )

    used = fields.Integer(
        metadata={
            'description': 'Used disk space in bytes',
            'example': 1024**2,
        }
    )

    free = fields.Integer(
        metadata={
            'description': 'Free disk space in bytes',
            'example': (1024**3) - (1024**2),
        }
    )

    read_count = fields.Integer(
        metadata={
            'description': 'Number of read operations',
            'example': 100,
        }
    )

    write_count = fields.Integer(
        metadata={
            'description': 'Number of write operations',
            'example': 50,
        }
    )

    read_bytes = fields.Integer(
        metadata={
            'description': 'Number of bytes read',
            'example': 1024**3,
        }
    )

    write_bytes = fields.Integer(
        metadata={
            'description': 'Number of bytes written',
            'example': 1024**2,
        }
    )

    read_time = fields.Float(
        metadata={
            'description': 'Time spent reading in seconds',
            'example': 10.5,
        }
    )

    write_time = fields.Float(
        metadata={
            'description': 'Time spent writing in seconds',
            'example': 5.5,
        }
    )

    busy_time = fields.Float(
        metadata={
            'description': 'Time spent doing I/Os in seconds',
            'example': 20.5,
        }
    )

    percent = percent_field(
        metadata={
            'description': 'Percentage of disk space used, normalized between 0 and 1',
            'example': 0.1,
        }
    )

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)

        # Convert read/write/busy times from milliseconds to seconds
        for attr in ['read_time', 'write_time', 'busy_time']:
            if data.get(attr) is not None:
                data[attr] /= 1000

        # Normalize the percentage between 0 and 1
        if data.get('percent') is not None:
            data['percent'] /= 100

        return data



from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import List, Optional


class LinodeInstanceStatus(Enum):
    """
    Maps the possible states of an instance.
    """

    RUNNING = 'running'
    OFFLINE = 'offline'
    BOOTING = 'booting'
    REBOOTING = 'rebooting'
    SHUTTING_DOWN = 'shutting_down'
    PROVISIONING = 'provisioning'
    DELETING = 'deleting'
    MIGRATING = 'migrating'
    REBUILDING = 'rebuilding'
    CLONING = 'cloning'
    RESTORING = 'restoring'
    STOPPED = 'stopped'


class LinodeInstanceBackupScheduleDay(Enum):
    """
    Allowed values for ``backups.schedule.day``.
    """

    SCHEDULING = 'Scheduling'
    SUNDAY = 'Sunday'
    MONDAY = 'Monday'
    TUESDAY = 'Tuesday'
    WEDNESDAY = 'Wednesday'
    THURSDAY = 'Thursday'
    FRIDAY = 'Friday'
    SATURDAY = 'Saturday'


class LinodeInstanceBackupScheduleWindow(Enum):
    """
    Allowed values for ``backups.schedule.window``.

    The window in which your backups will be taken, in UTC. A backups window is
    a two-hour span of time in which the backup may occur.

    For example, W10 indicates that your backups should be taken between 10:00
    and 12:00.
    """

    SCHEDULING = 'Scheduling'
    W0 = 'W0'
    W2 = 'W2'
    W4 = 'W4'
    W6 = 'W6'
    W8 = 'W8'
    W10 = 'W10'
    W12 = 'W12'
    W14 = 'W14'
    W16 = 'W16'
    W18 = 'W18'
    W20 = 'W20'
    W22 = 'W22'


@dataclass
class LinodeInstanceSpecs:
    """
    Class that models the specifications of a Linode instance.
    """

    disk: int
    memory: int
    cpus: int
    gpus: int
    transfer: int


@dataclass
class LinodeInstanceAlerts:
    """
    Class that models the alerts configuration of a Linode instance.
    """

    cpu: int
    io: int
    network_in: int
    network_out: int
    transfer_quota: int


@dataclass
class LinodeInstanceBackupSchedule:
    """
    Class that models the backup schedule of a Linode instance.
    """

    day: Optional[LinodeInstanceBackupScheduleDay]
    window: Optional[LinodeInstanceBackupScheduleWindow]


@dataclass
class LinodeInstanceBackups:
    """
    Class that models the backup status of a Linode instance.
    """

    available: bool
    enabled: bool
    schedule: LinodeInstanceBackupSchedule
    last_successful: Optional[datetime] = None


@dataclass
class LinodeInstance:
    """
    Class that models a Linode instance.
    """

    id: int
    name: str
    instance_type: str
    ipv4_addresses: List[str]
    ipv6_address: str
    group: str
    status: LinodeInstanceStatus
    tags: List[str]
    image: str
    region: str
    hypervisor: str
    specs: LinodeInstanceSpecs
    alerts: LinodeInstanceAlerts
    backups: LinodeInstanceBackups
    created_at: datetime
    updated_at: datetime



from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional


@dataclass
class Process:
    """
    System process data class.
    """

    pid: int
    name: str
    parent_pid: Optional[int] = None
    username: Optional[str] = None
    command_line: List[str] = field(default_factory=list)
    status: Optional[str] = None
    terminal: Optional[str] = None
    current_directory: Optional[str] = None
    started: Optional[datetime] = None
    cpu_percent: float = 0
    memory_percent: float = 0



from dataclasses import dataclass
from datetime import datetime
from typing import Optional


@dataclass
class User:
    """
    System user wrapper.
    """

    username: str
    terminal: Optional[str] = None
    started: Optional[datetime] = None
    pid: Optional[int] = None



from marshmallow import fields, pre_load

from platypush.schemas.dataclasses import percent_field
from .._base import SystemBaseSchema


class CpuInfoSchema(SystemBaseSchema):
    """
    Base schema for CPU info.
    """

    architecture = fields.String(
        data_key='arch_string_raw',
        metadata={
            'description': 'CPU architecture.',
            'example': 'x86_64',
        },
    )

    bits = fields.Int(
        metadata={
            'description': 'CPU architecture bits.',
            'example': 64,
        }
    )

    cores = fields.Int(
        data_key='count',
        metadata={
            'description': 'Number of CPU cores.',
            'example': 4,
        },
    )

    vendor = fields.String(
        data_key='vendor_id_raw',
        metadata={
            'description': 'CPU vendor.',
            'example': 'GenuineIntel',
        },
    )

    brand = fields.String(
        data_key='brand_raw',
        metadata={
            'description': 'CPU brand.',
            'example': 'Intel(R) Core(TM) i7-7700HQ CPU @ 2.80GHz',
        },
    )

    frequency_advertised = fields.Float(
        metadata={
            'description': 'CPU advertised frequency.',
            'example': 2800000000.0,
        }
    )

    frequency_actual = fields.Float(
        metadata={
            'description': 'CPU actual frequency.',
            'example': 2650000000.0,
        }
    )

    flags = fields.List(
        fields.String(),
        metadata={
            'description': 'CPU flags.',
            'example': ['fpu', 'vme', 'de', 'pse', 'tsc', 'msr', 'pae'],
        },
    )

    l1_instruction_cache_size = fields.Int(
        metadata={
            'description': 'L1 instruction cache size.',
            'example': 65536,
        }
    )

    l1_data_cache_size = fields.Int(
        metadata={
            'description': 'L1 data cache size.',
            'example': 65536,
        }
    )

    l2_cache_size = fields.Int(
        metadata={
            'description': 'L2 cache size.',
            'example': 524288,
        }
    )

    l3_cache_size = fields.Int(
        metadata={
            'description': 'L3 cache size.',
            'example': 6291456,
        }
    )

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)
        if data.get('hz_advertised'):
            data['frequency_advertised'] = data.pop('hz_advertised')[0]
        if data.get('hz_actual'):
            data['frequency_actual'] = data.pop('hz_actual')[0]

        for key, value in data.items():
            if key.endswith("_cache_size") and isinstance(value, str):
                tokens = value.split(" ")
                unit = None
                if len(tokens) > 1:
                    unit = tokens[1]

                value = int(tokens[0])
                if unit == "KiB":
                    value *= 1024
                elif unit == "MiB":
                    value *= 1024 * 1024
                elif unit == "GiB":
                    value *= 1024 * 1024 * 1024

                data[key] = value

        return data


class CpuTimesSchema(SystemBaseSchema):
    """
    Base schema for CPU times.
    """

    user = percent_field(
        metadata={
            'description': 'Time spent in user mode, '
            'as a percentage between 0 and 1 of the total CPU time.',
            'example': 0.0,
        }
    )

    nice = percent_field(
        metadata={
            'description': 'Time spent in nice mode, '
            'as a percentage between 0 and 1 of the total CPU time.',
            'example': 0.0,
        }
    )

    system = percent_field(
        metadata={
            'description': 'Time spent in system mode, '
            'as a percentage between 0 and 1 of the total CPU time.',
            'example': 0.0,
        }
    )

    idle = percent_field(
        metadata={
            'description': 'Time spent in idle mode, '
            'as a percentage between 0 and 1 of the total CPU time.',
            'example': 0.0,
        }
    )

    iowait = percent_field(
        metadata={
            'description': 'Time spent in I/O wait mode, '
            'as a percentage between 0 and 1 of the total CPU time.',
            'example': 0.0,
        }
    )

    irq = percent_field(
        metadata={
            'description': 'Time spent in IRQ mode, '
            'as a percentage between 0 and 1 of the total CPU time.',
            'example': 0.0,
        }
    )

    softirq = percent_field(
        metadata={
            'description': 'Time spent in soft IRQ mode, '
            'as a percentage between 0 and 1 of the total CPU time.',
            'example': 0.0,
        }
    )

    steal = percent_field(
        metadata={
            'description': 'Time spent in steal mode, '
            'as a percentage between 0 and 1 of the total CPU time.',
            'example': 0.0,
        }
    )

    guest = percent_field(
        metadata={
            'description': 'Time spent in guest mode, '
            'as a percentage between 0 and 1 of the total CPU time.',
            'example': 0.0,
        }
    )

    guest_nice = percent_field(
        metadata={
            'description': 'Time spent in guest nice mode, '
            'as a percentage between 0 and 1 of the total CPU time.',
            'example': 0.0,
        }
    )

    @pre_load
    def pre_load(self, data, **_) -> dict:
        """
        Convert the underlying object to dict and normalize all the percentage
        values from [0, 100] to [0, 1].
        """
        data = super().pre_load(data)
        return {
            key: value / 100.0
            for key, value in (
                data if isinstance(data, dict) else data._asdict()
            ).items()
        }


class CpuFrequencySchema(SystemBaseSchema):
    """
    Base schema for CPU frequency.
    """

    current = fields.Float(
        metadata={
            'description': 'Current CPU frequency.',
            'example': 2800000000.0,
        }
    )

    min = fields.Float(
        metadata={
            'description': 'Minimum CPU frequency.',
            'example': 800000000.0,
        }
    )

    max = fields.Float(
        metadata={
            'description': 'Maximum CPU frequency.',
            'example': 2800000000.0,
        }
    )


class CpuStatsSchema(SystemBaseSchema):
    """
    Base schema for CPU stats.
    """

    ctx_switches = fields.Int(
        metadata={
            'description': 'Number of context switches.',
            'example': 1234,
        }
    )

    interrupts = fields.Int(
        metadata={
            'description': 'Number of interrupts.',
            'example': 1234,
        }
    )

    soft_interrupts = fields.Int(
        metadata={
            'description': 'Number of soft interrupts.',
            'example': 1234,
        }
    )

    syscalls = fields.Int(
        metadata={
            'description': 'Number of system calls.',
            'example': 1234,
        }
    )


class CpuSchema(SystemBaseSchema):
    """
    Base schema for CPU results.
    """

    info = fields.Nested(CpuInfoSchema)
    times = fields.Nested(CpuTimesSchema)
    frequency = fields.Nested(CpuFrequencySchema)
    stats = fields.Nested(CpuStatsSchema)
    load_avg = fields.Tuple(
        [fields.Float(), fields.Float(), fields.Float()],
        metadata={
            'description': 'CPU load average.',
            'example': (0.0, 0.0, 0.0),
        },
    )

    percent = percent_field(
        metadata={
            'description': 'CPU usage percentage, as a value between 0 and 1.',
            'example': 0.0,
        }
    )



# Supported audio extensions
audio_extensions = {
    '3gp',
    'aa',
    'aac',
    'aax',
    'act',
    'aiff',
    'amr',
    'ape',
    'au',
    'awb',
    'dct',
    'dss',
    'dvf',
    'flac',
    'gsm',
    'iklax',
    'ivs',
    'm4a',
    'm4b',
    'm4p',
    'mmf',
    'mp3',
    'mpc',
    'msv',
    'nmf',
    'nsf',
    'ogg,',
    'opus',
    'ra,',
    'raw',
    'sln',
    'tta',
    'vox',
    'wav',
    'wma',
    'wv',
    'webm',
    '8svx',
}


# Supported video extensions
video_extensions = {
    'webm',
    'mkv',
    'flv',
    'vob',
    'ogv',
    'ogg',
    'drc',
    'gif',
    'gifv',
    'mng',
    'avi',
    'mts',
    'm2ts',
    'mov',
    'qt',
    'wmv',
    'yuv',
    'rm',
    'rmvb',
    'asf',
    'amv',
    'mp4',
    'm4p',
    'm4v',
    'mpg',
    'mp2',
    'mpeg',
    'mpe',
    'mpv',
    'm2v',
    'svi',
    '3gp',
    '3g2',
    'mxf',
    'roq',
    'nsv',
    'f4v',
    'f4p',
    'f4a',
    'f4b',
}


# vim:sw=4:ts=4:et:



import os
from typing import Optional

from ..._search.local.metadata import get_file_metadata
from .. import FileMediaResource
from ._base import MediaResourceParser


# pylint: disable=too-few-public-methods
class FileResourceParser(MediaResourceParser):
    """
    Parser for local file resources.
    """

    def parse(self, resource: str, *_, **__) -> Optional[FileMediaResource]:
        if resource.startswith('file://') or os.path.isfile(resource):
            path = resource
            if resource.startswith('file://'):
                path = resource[len('file://') :]

            assert os.path.isfile(path), f'File {path} not found'
            metadata = get_file_metadata(path)
            metadata['timestamp'] = metadata.pop('created_at', None)

            return FileMediaResource(
                resource=path,
                url=f'file://{path}',
                media_plugin=self._media,
                title=os.path.basename(resource),
                filename=os.path.basename(resource),
                **metadata,
            )

        return None


# vim:sw=4:ts=4:et:



from marshmallow import fields, pre_load

from .._base import SystemBaseSchema


class TemperatureSchema(SystemBaseSchema):
    """
    Schema for system temperature sensors.
    """

    id = fields.String(
        metadata={
            'description': 'The unique identifier of the temperature sensor.',
            'example': 'acpi_1',
        }
    )

    label = fields.String(
        allow_none=True,
        metadata={
            'description': 'The label of the temperature sensor.',
            'example': 'CPU Temperature',
        },
    )

    value = fields.Float(
        allow_none=True,
        metadata={
            'description': 'The current temperature value of the sensor, in degrees Celsius.',
            'example': 56.0,
        },
    )

    high = fields.Float(
        allow_none=True,
        metadata={
            'description': 'The high temperature threshold of the sensor, in degrees Celsius.',
            'example': 90.0,
        },
    )

    critical = fields.Float(
        allow_none=True,
        metadata={
            'description': 'The critical temperature threshold of the sensor, in degrees Celsius.',
            'example': 100.0,
        },
    )

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)
        data['value'] = data.pop('current', data.pop('value', None))
        return data



from dataclasses import dataclass
from socket import AddressFamily, SocketKind
from typing import Optional


@dataclass
class Connection:
    """
    Network/UNIX socket data class.
    """

    fd: int
    family: AddressFamily
    type: SocketKind
    local_address: str
    local_port: Optional[int] = None
    remote_address: Optional[str] = None
    remote_port: Optional[int] = None
    status: Optional[str] = None
    pid: Optional[int] = None



from ._base import MediaResource
from .file import FileMediaResource
from .http import HttpMediaResource
from .youtube import YoutubeMediaResource


__all__ = [
    'FileMediaResource',
    'HttpMediaResource',
    'MediaResource',
    'YoutubeMediaResource',
]



from ._base import DownloadThread, MediaResourceDownloader
from .http import HttpResourceDownloader
from .youtube import YoutubeResourceDownloader

downloaders = [
    YoutubeResourceDownloader,
    HttpResourceDownloader,
]

__all__ = [
    'DownloadThread',
    'HttpResourceDownloader',
    'MediaResourceDownloader',
    'YoutubeResourceDownloader',
    'downloaders',
]



import os
import pathlib
import queue
import tempfile
import threading
from abc import ABC, abstractmethod
from typing import (
    Dict,
    Iterable,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)
from urllib.parse import urlparse

import requests

from platypush.config import Config
from platypush.context import get_plugin, get_backend
from platypush.message.event.media import MediaEvent
from platypush.plugins import RunnablePlugin, action
from platypush.utils import (
    get_default_downloads_dir,
    get_mime_type,
    get_plugin_name_by_class,
)

from ._constants import audio_extensions, video_extensions
from ._model import DownloadState, MediaDirectory, PlayerState
from ._resource import MediaResource
from ._resource.downloaders import DownloadThread, MediaResourceDownloader, downloaders
from ._resource.parsers import MediaResourceParser, parsers
from ._search import MediaSearcher, searchers

_MediaDirs = Union[str, Iterable[Union[str, dict]], Dict[str, Union[str, dict]]]


class MediaPlugin(RunnablePlugin, ABC):
    """
    Generic plugin to interact with a media player.

    To start the local media stream service over HTTP you will also need the
    :class:`platypush.backend.http.HttpBackend` backend enabled.
    """

    # A media plugin can either be local or remote (e.g. control media on
    # another device)
    _is_local = True
    _NOT_IMPLEMENTED_ERR = NotImplementedError(
        'This method must be implemented in a derived class'
    )

    audio_extensions = audio_extensions
    video_extensions = video_extensions
    supported_media_plugins = [
        'media.vlc',
        'media.mpv',
        'media.gstreamer',
        'media.mplayer',
        'media.chromecast',
        'media.kodi',
    ]

    _supported_media_types = ['file', 'jellyfin', 'plex', 'torrent', 'youtube']
    _default_search_timeout = 60  # 60 seconds

    def __init__(
        self,
        media_dirs: Optional[_MediaDirs] = None,
        download_dir: Optional[str] = None,
        env: Optional[Dict[str, str]] = None,
        volume: Optional[Union[float, int]] = None,
        torrent_plugin: str = 'torrent',
        youtube_format: Optional[str] = 'bv[height<=?1080]+ba/bv+ba',
        youtube_audio_format: Optional[str] = 'ba',
        youtube_dl: str = 'yt-dlp',
        merge_output_format: str = 'mp4',
        cache_dir: Optional[str] = None,
        cache_streams: bool = False,
        ytdl_args: Optional[Sequence[str]] = None,
        **kwargs,
    ):
        """
        :param media_dirs: Directories that will be scanned for media files when
            a search is performed (default: only ``download_dir``). You can
            specify it either:

                - As a list of strings:

                  .. code-block:: yaml

                    media_dirs:
                        - /mnt/hd/media/movies
                        - /mnt/hd/media/music
                        - /mnt/hd/media/series

                - As a dictionary where the key is the name of the media display
                  name and the value is the path:

                    .. code-block:: yaml

                      media_dirs:
                          Movies: /mnt/hd/media/movies
                          Music: /mnt/hd/media/music
                          Series: /mnt/hd/media/series

                - As a dictionary where the key is the name of the media display
                  name and the value is a dictionary with the path and additional
                  display information:

                      media_dirs:
                          Movies:
                              path: /mnt/hd/media/movies
                              icon:
                                  url: https://example.com/icon.png
                                  # FontAwesome icon classes are supported
                                  class: fa fa-film

                          Music:
                              path: /mnt/hd/media/music
                              icon:
                                  url: https://example.com/icon.png
                                  class: fa fa-music

                          Series:
                              path: /mnt/hd/media/series
                              icon:
                                  url: https://example.com/icon.png
                                  class: fa fa-tv

        :param download_dir: Directory where external resources/torrents will be
            downloaded (default: ~/Downloads)
        :param env: Environment variables key-values to pass to the
            player executable (e.g. DISPLAY, XDG_VTNR, PULSE_SINK etc.)
        :param volume: Default volume for the player (default: None, maximum volume).
        :param torrent_plugin: Optional plugin to be used for torrent download.
            Possible values:

                - ``torrent`` - native ``libtorrent``-based plugin (default,
                  recommended)
                - ``rtorrent`` - torrent support over rtorrent RPC/XML interface
                - ``webtorrent`` - torrent support over webtorrent (unstable)

        :param youtube_format: Select the preferred video/audio format for
            YouTube videos - and any media supported by youtube-dl or the
            selected fork. See the `youtube-dl documentation
            <https://github.com/ytdl-org/youtube-dl#format-selection>`_ for more
            info on supported formats. Example:
            ``bestvideo[height<=?1080][ext=mp4]+bestaudio`` - select the best
            mp4 video with a resolution <= 1080p, and the best audio format.
        :param youtube_audio_format: Select the preferred audio format for
            YouTube videos downloaded only for audio. Default: ``bestaudio``.
        :param youtube_dl: Path to the ``youtube-dl`` executable, used to
            extract information from YouTube videos and other media platforms.
            Default: ``yt-dlp``. The default has changed from ``youtube-dl`` to
            the ``yt-dlp`` fork because the former is badly maintained and its
            latest release was pushed in 2021.
        :param merge_output_format: If media download requires ``youtube_dl``,
            and the upstream media contains both audio and video to be merged,
            this can be used to specify the format of the output container -
            e.g. ``mp4``, ``mkv``, ``avi``, ``flv``. Default: ``mp4``.
        :param cache_dir: Directory where the media cache will be stored. If not
            specified, the cache will be stored in the default cache directory
            (usually ``~/.cache/platypush/media/<media_plugin>``).
        :param cache_streams: If set to True, streams transcoded via yt-dlp or
            ffmpeg will be cached in ``cache_dir`` directory. If not set
            (default), then streams will be played directly via memory pipe.
            You may want to set this to True if you have a slow network, or if
            you want to play media at high quality, even though the start time
            may be delayed. If set to False, the media will start playing as
            soon as the stream is ready, but the quality may be lower,
            especially at the beginning, and seeking may not be supported.
        :param ytdl_args: Additional arguments to pass to the youtube-dl
            executable. Default: None.
        """

        super().__init__(**kwargs)

        player = None
        player_config = {}
        self._download_threads: Dict[Tuple[str, str], DownloadThread] = {}

        if self.__class__.__name__ == 'MediaPlugin':
            # Abstract class, initialize with the default configured player
            for plugin_name in Config.get_plugins().keys():
                if plugin_name in self.supported_media_plugins:
                    player = get_plugin(plugin_name)
                    if player and player.is_local():
                        # Local players have priority as default if configured
                        break
        else:
            player = self  # Derived concrete class

        if not player:
            raise AttributeError('No media plugin configured')

        if self.__class__.__name__ == 'MediaPlugin':
            # Populate this plugin with the actions of the configured player
            for act in player.registered_actions:
                setattr(self, act, getattr(player, act))
                self.registered_actions.add(act)

        self._env = env or {}
        self.cache_streams = cache_streams
        self.download_dir = os.path.abspath(
            os.path.expanduser(
                download_dir
                or player_config.get('download_dir')
                or get_default_downloads_dir()
            )
        )

        self.cache_dir = os.path.abspath(
            os.path.expanduser(cache_dir)
            if cache_dir
            else os.path.join(
                Config.get_cachedir(),
                'media',
                get_plugin_name_by_class(self.__class__),
            )
        )

        pathlib.Path(self.cache_dir).mkdir(parents=True, exist_ok=True)
        pathlib.Path(self.download_dir).mkdir(parents=True, exist_ok=True)
        self._ytdl = youtube_dl
        self.volume = volume
        self._videos_queue = []
        self._youtube_proc = None
        self.torrent_plugin = torrent_plugin
        self.youtube_format = youtube_format
        self.youtube_audio_format = youtube_audio_format
        self.merge_output_format = merge_output_format
        self.ytdl_args = ytdl_args or []
        self._latest_resource: Optional[MediaResource] = None

        self.media_dirs = self._parse_media_dirs(
            media_dirs or player_config.get('media_dirs', [])
        )

        self._parsers: Dict[Type[MediaResourceParser], MediaResourceParser] = {
            parser: parser(self) for parser in parsers
        }

        self._downloaders: Dict[
            Type[MediaResourceDownloader], MediaResourceDownloader
        ] = {downloader: downloader(self) for downloader in downloaders}

        self._searchers: Dict[Type[MediaSearcher], MediaSearcher] = {
            searcher: searcher(
                dirs=[d.path for d in self.media_dirs.values()], media_plugin=self
            )
            for searcher in searchers
        }

    @staticmethod
    def _parse_media_dirs(
        media_dirs: Optional[_MediaDirs],
    ) -> Dict[str, MediaDirectory]:
        dirs = {}

        if media_dirs:
            if isinstance(media_dirs, str):
                dirs = [media_dirs]
            if isinstance(media_dirs, (list, tuple, set)):
                dirs = {d: d for d in media_dirs}
            if isinstance(media_dirs, dict):
                dirs = media_dirs

        assert isinstance(dirs, dict), f'Invalid media_dirs format: {media_dirs}'

        ret = {}
        for k, v in dirs.items():
            assert isinstance(k, str), f'Invalid media_dirs key format: {k}'
            if isinstance(v, str):
                v = {'path': v}

            assert isinstance(v, dict), f'Invalid media_dirs format: {v}'
            path = v.get('path')
            assert path, f'Missing path in media_dirs entry {k}'
            path = os.path.abspath(os.path.expanduser(path))
            assert os.path.isdir(path), f'Invalid path in media_dirs entry {k}'

            icon = v.get('icon', {})
            if isinstance(icon, str):
                # Fill up the URL field if it's a URL, otherwise assume that
                # it's a FontAwesome icon class
                icon = {'url': icon} if urlparse(icon).scheme else {'class': icon}

            ret[k] = MediaDirectory.build(
                name=k,
                path=path,
                icon_class=icon.get('class'),
                icon_url=icon.get('url'),
            )

        # Add the downloads directory if it's missing
        if not any(d.path == get_default_downloads_dir() for d in ret.values()):
            ret['Downloads'] = MediaDirectory.build(
                name='Downloads',
                path=get_default_downloads_dir(),
                icon_class='fas fa-download',
            )

        return {k: ret[k] for k in sorted(ret.keys())}

    def _get_resource(
        self,
        resource: str,
        metadata: Optional[dict] = None,
        only_audio: bool = False,
        **_,
    ):
        """
        :param resource: Resource to play/parse. Supported types:

            * Local files (format: ``file://<path>/<file>``)
            * Remote videos (format: ``https://<url>/<resource>``)
            * Torrents (format: Magnet links, Torrent URLs or local Torrent files)
            * Any URL that is supported by a yt_dlp extractor

        """

        for parser in self._parsers.values():
            media_resource = parser.parse(resource, only_audio=only_audio)
            if media_resource:
                for k, v in (metadata or {}).items():
                    setattr(media_resource, k, v)

                return media_resource

        raise AssertionError(f'Unknown media resource: {resource}')

    @action
    @abstractmethod
    def play(self, resource: str, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def pause(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def stop(self, *args, **kwargs):  # type: ignore
        super().stop()

    @action
    @abstractmethod
    def quit(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def voldown(self, step: Optional[float] = 5.0, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def volup(self, step: Optional[float] = 5.0, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def back(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def forward(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def next(self):
        """Play the next item in the queue"""
        self.stop()

        if self._videos_queue:
            video = self._videos_queue.pop(0)
            return self.play(video)

        return None

    @action
    @abstractmethod
    def toggle_subtitles(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def set_subtitles(self, filename, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def remove_subtitles(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def is_playing(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def load(self, resource, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def mute(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def seek(self, position, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def set_position(self, position, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def set_volume(self, volume):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def status(self):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def search(
        self,
        query: str,
        types: Optional[Iterable[str]] = None,
        queue_results: bool = False,
        autoplay: bool = False,
        timeout: float = _default_search_timeout,
    ):
        """
        Perform a video search.

        :param query: Query string, video name or partial name
        :param types: Video types to search (default: ``["youtube", "file", "torrent"]``)
        :param queue_results: Append the results to the current playing queue (default: False)
        :param autoplay: Play the first result of the search (default: False)
        :param timeout: Search timeout (default: 60 seconds)
        """

        results = {}
        results_queues = {}
        worker_threads = {}

        if types is None:
            types = self._supported_media_types

        for media_type in types:
            results[media_type] = []
            results_queues[media_type] = queue.Queue()
            search_hndl = self._get_search_handler_by_type(media_type)
            worker_threads[media_type] = threading.Thread(
                target=self._search_worker(
                    query=query,
                    search_hndl=search_hndl,
                    results_queue=results_queues[media_type],
                )
            )
            worker_threads[media_type].start()

        for media_type in types:
            try:
                items = results_queues[media_type].get(timeout=timeout)
                if isinstance(items, Exception):
                    raise items

                results[media_type].extend(items)
            except queue.Empty:
                self.logger.warning(
                    'Search for "%s" media type %s timed out', query, media_type
                )
            except Exception as e:
                self.logger.warning(
                    'Error while searching for "%s", media type %s', query, media_type
                )
                self.logger.exception(e)

        results = [
            {**result, 'type': media_type}
            for media_type in self._supported_media_types
            for result in results.get(media_type, [])
            if media_type in results
        ]

        if results:
            if queue_results:
                self._videos_queue = [_['url'] for _ in results]
                if autoplay:
                    self.play(self._videos_queue.pop(0))
            elif autoplay:
                self.play(results[0]['url'])

        return results

    @staticmethod
    def _search_worker(query, search_hndl, results_queue):
        def thread():
            try:
                results_queue.put(search_hndl.search(query))
            except Exception as e:
                results_queue.put(e)

        return thread

    def _get_search_handler_by_type(self, search_type: str):
        searcher = next(
            iter(filter(lambda s: s.supports(search_type), self._searchers.values())),
            None,
        )

        if not searcher:
            self.logger.warning('Unsupported search type: %s', search_type)
            return None

        return searcher

    @classmethod
    def is_video_file(cls, filename: str):
        if filename.lower().split('.')[-1] in cls.video_extensions:
            return True

        mime_type = get_mime_type(filename)
        return bool(mime_type and mime_type.startswith('video/'))

    @classmethod
    def is_audio_file(cls, filename: str):
        if filename.lower().split('.')[-1] in cls.audio_extensions:
            return True

        mime_type = get_mime_type(filename)
        return bool(mime_type and mime_type.startswith('audio/'))

    @classmethod
    def is_media_file(cls, file: str) -> bool:
        if file.split('.')[-1].lower() in cls.video_extensions.union(
            cls.audio_extensions
        ):
            return True

        mime_type = get_mime_type(file)
        return bool(
            mime_type
            and (mime_type.startswith('video/') or mime_type.startswith('audio/'))
        )

    @action
    def start_streaming(
        self, media: str, subtitles: Optional[str] = None, download: bool = False
    ):
        """
        Starts streaming local media over the specified HTTP port.
        The stream will be available to HTTP clients on
        `http://{this-ip}:{http_backend_port}/media/<media_id>`

        :param media: Media to stream
        :param subtitles: Path or URL to the subtitles track to be used
        :param download: Set to True if you prefer to download the file from
            the streaming link instead of streaming it
        :return: dict containing the streaming URL.Example:

        .. code-block:: json

            {
                "id": "0123456abcdef.mp4",
                "source": "file:///mnt/media/movies/movie.mp4",
                "mime_type": "video/mp4",
                "url": "http://192.168.1.2:8008/media/0123456abcdef.mp4"
            }

        """
        return self._start_streaming(media, subtitles=subtitles, download=download)

    def _start_streaming(
        self, media: str, subtitles: Optional[str] = None, download: bool = False
    ):
        http = get_backend('http')
        assert http, f'Unable to stream {media}: HTTP backend not configured'

        self.logger.info('Starting streaming %s', media)
        response = requests.put(
            f'{http.local_base_url}/media' + ('?download' if download else ''),
            json={'source': media, 'subtitles': subtitles},
            timeout=300,
        )

        assert response.ok, response.text or response.reason
        return response.json()

    @action
    def stop_streaming(self, media_id: str):
        http = get_backend('http')
        assert http, f'Unable to stop streaming {media_id}: HTTP backend not configured'

        response = requests.delete(
            f'{http.local_base_url}/media/{media_id}', timeout=30
        )

        assert response.ok, response.text or response.reason
        return response.json()

    @action
    def get_info(self, resource: str):
        for parser in self._parsers.values():
            info = parser.parse(resource)
            if info:
                return info.to_dict()

        return {'url': resource}

    @action
    def download(
        self,
        url: str,
        filename: Optional[str] = None,
        directory: Optional[str] = None,
        timeout: int = 10,
        sync: bool = False,
        only_audio: bool = False,
        youtube_format: Optional[str] = None,
        youtube_audio_format: Optional[str] = None,
        merge_output_format: Optional[str] = None,
    ):
        """
        Download a media URL to a local file on the Platypush host (yt-dlp
        required for YouTube URLs).

        This action is non-blocking and returns the path to the downloaded file
        once the download is initiated.

        You can then subscribe to these events to monitor the download progress:

            - :class:`platypush.message.event.media.MediaDownloadStartedEvent`
            - :class:`platypush.message.event.media.MediaDownloadProgressEvent`
            - :class:`platypush.message.event.media.MediaDownloadErrorEvent`
            - :class:`platypush.message.event.media.MediaDownloadPausedEvent`
            - :class:`platypush.message.event.media.MediaDownloadResumedEvent`
            - :class:`platypush.message.event.media.MediaDownloadCancelledEvent`
            - :class:`platypush.message.event.media.MediaDownloadCompletedEvent`

        :param url: Media URL.
        :param filename: Media filename (default: inferred from the URL basename).
        :param directory: Destination directory (default: ``download_dir``).
        :param timeout: Network timeout in seconds (default: 10).
        :param sync: If set to True, the download will be synchronous and the
            action will return only when the download is completed.
        :param only_audio: If set to True, only the audio track will be downloaded
            (only supported for yt-dlp-compatible URLs for now).
        :param youtube_format: Override the default ``youtube_format`` setting.
        :param youtube_audio_format: Override the default ``youtube_audio_format``
        :param merge_output_format: Override the default
            ``merge_output_format`` setting.
        :return: The absolute path to the downloaded file.
        """
        dl_thread = None
        resource = self._get_resource(url, only_audio=only_audio)

        if filename:
            path = os.path.expanduser(filename)
        elif resource.filename:
            path = resource.filename
        else:
            path = os.path.basename(resource.url)

        if not os.path.isabs(path):
            directory = os.path.expanduser(directory or self.download_dir)
            path = os.path.join(directory, path)

        for downloader in self._downloaders.values():
            if downloader.supports(resource):
                if only_audio and not downloader.supports_only_audio():
                    self.logger.warning(
                        'Only audio download is not supported for this resource'
                    )

                dl_thread = downloader.download(
                    resource=resource,
                    path=path,
                    timeout=timeout,
                    only_audio=only_audio,
                    youtube_format=youtube_format or self.youtube_format,
                    youtube_audio_format=youtube_audio_format
                    or self.youtube_audio_format,
                    merge_output_format=merge_output_format,
                )

                break

        assert dl_thread, f'No downloader found for resource {url}'

        if sync:
            dl_thread.join()

        return path

    @action
    def pause_download(self, url: Optional[str] = None, path: Optional[str] = None):
        """
        Pause a download in progress.

        Either the URL or the path must be specified.

        :param url: URL of the download.
        :param path: Path of the download (default: any path associated with the URL).
        """
        for thread in self._get_downloads(url=url, path=path):
            thread.pause()

    @action
    def resume_download(self, url: Optional[str] = None, path: Optional[str] = None):
        """
        Resume a paused download.

        Either the URL or the path must be specified.

        :param url: URL of the download.
        :param path: Path of the download (default: any path associated with the URL).
        """
        for thread in self._get_downloads(url=url, path=path):
            thread.resume()

    @action
    def cancel_download(self, url: Optional[str] = None, path: Optional[str] = None):
        """
        Cancel a download in progress.

        Either the URL or the path must be specified.

        :param url: URL of the download.
        :param path: Path of the download (default: any path associated with the URL).
        """
        for thread in self._get_downloads(url=url, path=path):
            thread.stop()

    @action
    def clear_downloads(self, url: Optional[str] = None, path: Optional[str] = None):
        """
        Clear completed/cancelled downloads from the queue.

        :param url: URL of the download (default: all downloads).
        :param path: Path of the download (default: any path associated with the URL).
        """
        threads = (
            self._get_downloads(url=url, path=path)
            if url
            else list(self._download_threads.values())
        )

        for thread in threads:
            if thread.state not in (DownloadState.COMPLETED, DownloadState.CANCELLED):
                continue

            dl = self._download_threads.pop((thread.url, thread.path), None)
            if dl:
                dl.clear()

    @action
    def get_downloads(self, url: Optional[str] = None, path: Optional[str] = None):
        """
        Get the download threads.

        :param url: URL of the download (default: all downloads).
        :param path: Path of the download (default: any path associated with the URL).
        :return: .. schema:: media.download.MediaDownloadSchema(many=True)
        """
        from platypush.schemas.media.download import MediaDownloadSchema

        return MediaDownloadSchema().dump(
            (
                self._get_downloads(url=url, path=path)
                if url
                else list(self._download_threads.values())
            ),
            many=True,
        )

    @action
    def get_media_dirs(self) -> Dict[str, dict]:
        """
        :return: List of configured media directories.
        """
        return {dir.name: dir.to_dict() for dir in self.media_dirs.values()}

    def _get_downloads(self, url: Optional[str] = None, path: Optional[str] = None):
        assert url or path, 'URL or path must be specified'
        threads = []

        if url and path:
            path = os.path.expanduser(path)
            thread = self._download_threads.get((url, path))
            if thread:
                threads = [thread]
        elif url:
            threads = [
                thread
                for (url_, _), thread in self._download_threads.items()
                if url_ == url
            ]
        elif path:
            path = os.path.expanduser(path)
            threads = [
                thread
                for (_, path_), thread in self._download_threads.items()
                if path_ == path
            ]

        assert threads, f'No matching downloads found for [url={url}, path={path}]'
        return threads

    def on_download_start(self, thread: DownloadThread):
        self._download_threads[thread.url, thread.path] = thread

    def start_download(self, thread: DownloadThread):
        if (thread.url, thread.path) in self._download_threads:
            self.logger.warning(
                'A download of %s to %s is already in progress', thread.url, thread.path
            )
            return

        thread.start()

    def post_event(self, event_type: Type[MediaEvent], **kwargs):
        evt = event_type(
            player=get_plugin_name_by_class(self.__class__), plugin=self, **kwargs
        )
        self._bus.post(evt)

    def is_local(self):
        return self._is_local

    @staticmethod
    def get_subtitles_file(subtitles: Optional[str] = None):
        if not subtitles:
            return None

        if subtitles.startswith('file://'):
            subtitles = subtitles[len('file://') :]
        if os.path.isfile(subtitles):
            return os.path.abspath(subtitles)

        content = requests.get(subtitles, timeout=20).content
        f = tempfile.NamedTemporaryFile(
            prefix='media_subs_', suffix='.srt', delete=False
        )

        with f:
            f.write(content)
        return f.name

    @property
    def supports_local_media(self) -> bool:
        return True

    @property
    def supports_local_pipe(self) -> bool:
        return True

    def main(self):
        self.wait_stop()


__all__ = [
    'DownloadState',
    'MediaPlugin',
    'MediaResource',
    'MediaSearcher',
    'PlayerState',
]


# vim:sw=4:ts=4:et:



from datetime import datetime

from dateutil.tz import gettz
from marshmallow import fields, pre_load

from platypush.schemas import DateTime

from .._base import SystemBaseSchema


class UserSchema(SystemBaseSchema):
    """
    Schema for system users.
    """

    username = fields.String(
        required=True,
        metadata={'description': 'The username of the user.', 'example': 'johndoe'},
    )

    terminal = fields.String(
        metadata={
            'description': 'The terminal the user is currently using.',
            'example': 'tty1',
        }
    )

    started = DateTime(
        metadata={
            'description': 'The timestamp when the user session started.',
            'example': '2021-01-01T00:00:00+00:00',
        }
    )

    pid = fields.Integer(
        metadata={'description': 'The PID of the user session.', 'example': 1234}
    )

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)
        started_ts = data.pop('started', None)
        if started_ts is not None:
            data['started'] = datetime.fromtimestamp(started_ts).replace(tzinfo=gettz())

        data['username'] = data.pop('name', data.pop('username', None))
        return data



from dataclasses import dataclass
from typing import IO

from ._base import MediaResource


@dataclass
class FileMediaResource(MediaResource):
    """
    Models a media resource that is read from a file.
    """

    def open(self, *args, **kwargs) -> IO:
        """
        Opens the media resource.
        """
        if self.fd is None:
            self.fd = open(self.resource, 'rb')  # pylint: disable=consider-using-with

        return super().open(*args, **kwargs)


# vim:sw=4:ts=4:et:



import os
from typing import Optional

from .. import HttpMediaResource
from ._base import MediaResourceParser


# pylint: disable=too-few-public-methods
class HttpResourceParser(MediaResourceParser):
    """
    Parser for HTTP resources.
    """

    def parse(self, resource: str, *_, **__) -> Optional[HttpMediaResource]:
        if resource.startswith('http://') or resource.startswith('https://'):
            assert self._media.is_media_file(
                resource
            ), f'Invalid media resource: {resource}'

            return HttpMediaResource(
                resource=resource,
                url=resource,
                media_plugin=self._media,
                title=os.path.basename(resource),
                filename=os.path.basename(resource),
            )

        return None


# vim:sw=4:ts=4:et:
